@@ -1,6 +1,6 @@
 /* 
    HTTP request handling tests
-   Copyright (C) 2001-2007, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2001-2009, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -203,7 +203,7 @@ static int reason_phrase(void)
 
     CALL(make_session(&sess, single_serve_string, RESP200
 		      "Connection: close\r\n\r\n"));
-    CALL(any_request(sess, "/foo"));
+    ONREQ(any_request(sess, "/foo"));
     CALL(await_server());
     
     ONV(strcmp(ne_get_error(sess), "200 OK"),
@@ -227,7 +227,7 @@ static int single_get_clength(void)
 {
     return expect_response("a", single_serve_string,
 			   RESP200
-			   "Content-Length: 1\r\n"
+			   "Content-Length: \t\t 1 \t\t\r\n"
 			   "\r\n"
 			   "a"
 			   "bbbbbbbbasdasd");
@@ -262,18 +262,6 @@ static int no_body_HEAD(void)
 			  "Content-Length: 5\r\n\r\n");
 }
 
-static int no_body_empty_clength(void)
-{
-    return expect_no_body("GET", "HTTP/1.1 200 OK\r\n"
-			  "Content-Length:\r\n\r\n");
-}
-
-static int no_body_bad_clength(void)
-{
-    return expect_no_body("GET", "HTTP/1.1 200 OK\r\n"
-			  "Content-Length: foobar\r\n\r\n");
-}
-
 static int no_headers(void)
 {
     return expect_response("abcde", single_serve_string,
@@ -421,14 +409,19 @@ static int serve_twice(ne_socket *sock, void *userdata)
 /* Test persistent connection handling: serve 'response' twice on a
  * single TCP connection, expecting to get a response body equal to
  * 'body' both times. */
-static int test_persist(const char *response, const char *body)
+static int test_persist_p(const char *response, const char *body, int proxy)
 {
     ne_session *sess = ne_session_create("http", "localhost", 7777);
     ne_buffer *buf = ne_buffer_create();
 
     ON(sess == NULL || buf == NULL);
     ON(spawn_server(7777, serve_twice, (char *)response));
-    
+
+    if (proxy) {
+        ne_session_proxy(sess, "localhost", 7777);
+        ne_set_session_flag(sess, NE_SESSFLAG_CONNAUTH, 1);
+    }
+
     CALL(run_request(sess, 200, construct_get, buf));
     
     ONV(strcmp(buf->data, body),
@@ -449,6 +442,11 @@ static int test_persist(const char *response, const char *body)
     return OK;
 }
 
+static int test_persist(const char *response, const char *body)
+{
+    return test_persist_p(response, body, 0);
+}
+
 static int persist_http11(void)
 {
     return test_persist(RESP200 "Content-Length: 5\r\n\r\n" "abcde",
@@ -469,6 +467,14 @@ static int persist_http10(void)
 			"abcde");
 }
 
+static int persist_proxy_http10(void)
+{
+    return test_persist_p("HTTP/1.0 200 OK\r\n"
+                          "Proxy-Connection: keep-alive\r\n"
+                          "Content-Length: 5\r\n\r\n" "abcde",
+                          "abcde", 1);
+}
+
 /* Server function for fail_early_eof */
 static int serve_eof(ne_socket *sock, void *ud)
 {
@@ -1581,6 +1587,15 @@ static int fail_on_invalid(void)
         /* negative C-L */
         { RESP200 "Content-Length: -1\r\n" "\r\n" "abcde",
           "Invalid Content-Length" },
+
+        /* invalid C-Ls */
+        { RESP200 "Content-Length: 5, 3\r\n" "\r\n" "abcde",
+          "Invalid Content-Length" },
+        { RESP200 "Content-Length: 5z\r\n" "\r\n" "abcde",
+          "Invalid Content-Length" },
+        { RESP200 "Content-Length: z5\r\n" "\r\n" "abcde",
+          "Invalid Content-Length" },
+
         /* stupidly-large C-L */
         { RESP200 "Content-Length: 99999999999999999999999999\r\n" 
           "\r\n" "abcde",
@@ -1607,12 +1622,12 @@ static int versions(void)
 		      "HTTP/1.0 200 OK\r\n"
 		      "Content-Length: 0\r\n\r\n"));
     
-    CALL(any_request(sess, "/http11"));
+    ONREQ(any_request(sess, "/http11"));
 	 
     ONN("did not detect HTTP/1.1 compliance",
 	ne_version_pre_http11(sess) != 0);
 	 
-    CALL(any_request(sess, "/http10"));
+    ONREQ(any_request(sess, "/http10"));
 
     ONN("did not detect lack of HTTP/1.1 compliance",
 	ne_version_pre_http11(sess) == 0);
@@ -1655,7 +1670,7 @@ static int hook_create_req(void)
     args.uri = "/foo";
     args.result = -1;
 
-    CALL(any_request(sess, "/foo"));
+    ONREQ(any_request(sess, "/foo"));
     
     ONN("first hook never called", args.result == -1);
     if (args.result) return FAIL;
@@ -1666,7 +1681,7 @@ static int hook_create_req(void)
     /* force use of absoluteURI in request-uri */
     ne_session_proxy(sess, "localhost", 7777);
 
-    CALL(any_request(sess, "/bar"));
+    ONREQ(any_request(sess, "/bar"));
     
     ONN("second hook never called", args.result == -1);
     if (args.result) return FAIL;
@@ -1757,24 +1772,22 @@ static int send_bad_offset(void)
 {
     ne_session *sess;
     ne_request *req;
-    char *fn = "empty.txt";
-    int ret, fd;
+    int ret, fds[2];
 
     CALL(make_session(&sess, single_serve_string,
                       RESP200 "Content-Length: 0\r\n" "\r\n"));
 
-    fd = open(fn, O_RDWR|O_CREAT);
-
-    ONV(fd < 0, ("could not find %s", fn));
+    /* create a pipe, on which seek is guaranteed to fail. */
+    ONN("could not create pipe", pipe(fds) != 0);
 
     req = ne_request_create(sess, "PUT", "/null");
 
-    ne_set_request_body_fd(req, fd, -500, 5);
+    ne_set_request_body_fd(req, fds[0], 500, 5);
     
     ret = ne_request_dispatch(req);
 
-    close(fd);
-    unlink(fn);
+    close(fds[0]);
+    close(fds[1]);
 
     ONN("request dispatched with bad offset!", ret == NE_OK);
     ONV(ret != NE_ERROR,
@@ -1853,6 +1866,13 @@ static void hook_destroy_sess(void *userdata)
     ne_buffer_czappend(buf, "(destroy-sess)\n");
 }
 
+static void hook_close_conn(void *userdata)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_czappend(buf, "(close-conn)\n");
+}
+
 static int hooks(void)
 {
     ne_buffer *buf = ne_buffer_create();
@@ -1870,6 +1890,7 @@ static int hooks(void)
     ne_hook_post_send(sess, hook_post_send, buf);
     ne_hook_destroy_request(sess, hook_destroy_req, buf);
     ne_hook_destroy_session(sess, hook_destroy_sess, buf);
+    ne_hook_close_conn(sess, hook_close_conn, buf);
 
     CALL(any_2xx_request(sess, "/first"));
 
@@ -1915,7 +1936,8 @@ static int hooks(void)
     ne_session_destroy(sess);
     CALL(await_server());
 
-    ONCMP("(destroy-sess)\n", buf->data, "hook ordering", "first destroyed session");
+    ONCMP("(destroy-sess)\n"
+          "(close-conn)\n", buf->data, "hook ordering", "first destroyed session");
 
     ne_buffer_clear(buf);
 
@@ -2010,6 +2032,23 @@ static int status(void)
 {
     ne_session *sess;
     ne_buffer *buf = ne_buffer_create();
+    ne_sock_addr *sa = ne_addr_resolve("localhost", 0);
+    char addr[64], expect[1024];
+
+    ONN("could not resolve localhost", ne_addr_result(sa));
+
+    ne_snprintf(expect, sizeof expect,
+                "lookup(localhost)-"
+                "connecting(localhost,%s)-"
+                "connected(localhost)-"
+                "send(0,5000)-"
+                "send(5000,5000)-"
+                "recv(0,5)-"
+                "recv(5,5)-"
+                "disconnected(localhost)-",
+                ne_iaddr_print(ne_addr_first(sa), addr, sizeof addr));
+
+    ne_addr_destroy(sa);
 
     CALL(make_session(&sess, single_serve_string, RESP200
                       "Content-Length: 5\r\n\r\n" "abcde"));
@@ -2021,15 +2060,9 @@ static int status(void)
     ne_session_destroy(sess);
     CALL(await_server());
 
-    ONCMP("lookup(localhost)-"
-          "connecting(localhost,127.0.0.1)-"
-          "connected(localhost)-"
-          "send(0,5000)-"
-          "send(5000,5000)-"
-          "recv(0,5)-"
-          "recv(5,5)-"
-          "disconnected(localhost)-",
-          buf->data, "status events", "result");
+    ONV(strcmp(expect, buf->data),
+        ("status event sequence mismatch: got [%s] not [%s]",
+         buf->data, expect));
     
     ne_buffer_destroy(buf);
     
@@ -2040,6 +2073,29 @@ static int status_chunked(void)
 {
     ne_session *sess;
     ne_buffer *buf = ne_buffer_create();
+    ne_sock_addr *sa = ne_addr_resolve("localhost", 0);
+    char addr[64], expect[1024];
+
+    ONN("could not resolve localhost", ne_addr_result(sa));
+
+    /* This sequence is not exactly guaranteed by the API, but it's
+     * what the current implementation should do. */
+    ne_snprintf(expect, sizeof expect,
+                "lookup(localhost)-"
+                "connecting(localhost,%s)-"
+                "connected(localhost)-"
+                "send(0,5000)-"
+                "send(5000,5000)-"
+                "recv(0,-1)-"
+                "recv(1,-1)-"
+                "recv(2,-1)-"
+                "recv(3,-1)-"
+                "recv(4,-1)-"
+                "recv(5,-1)-"
+                "disconnected(localhost)-",
+                ne_iaddr_print(ne_addr_first(sa), addr, sizeof addr));
+
+    ne_addr_destroy(sa);
 
     CALL(make_session(&sess, single_serve_string, 
                       RESP200 TE_CHUNKED "\r\n" ABCDE_CHUNKS));
@@ -2051,27 +2107,55 @@ static int status_chunked(void)
     ne_session_destroy(sess);
     CALL(await_server());
     
-    /* This sequence is not guaranted by the API, but it's what the
-     * current implementation should do. */
-    ONCMP("lookup(localhost)-"
-          "connecting(localhost,127.0.0.1)-"
-          "connected(localhost)-"
-          "send(0,5000)-"
-          "send(5000,5000)-"
-          "recv(0,-1)-"
-          "recv(1,-1)-"
-          "recv(2,-1)-"
-          "recv(3,-1)-"
-          "recv(4,-1)-"
-          "recv(5,-1)-"
-          "disconnected(localhost)-",
-          buf->data, "status events", "result");
+    ONV(strcmp(expect, buf->data),
+        ("status event sequence mismatch: got [%s] not [%s]",
+         buf->data, expect));
 
     ne_buffer_destroy(buf);
         
     return OK;
 }
 
+static const unsigned char raw_127[4] = "\x7f\0\0\01"; /* 127.0.0.1 */
+
+static int local_addr(void)
+{
+    ne_session *sess;
+    ne_inet_addr *ia = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
+
+    CALL(make_session(&sess, single_serve_string, RESP200 
+                      "Connection: close\r\n\r\n"));
+
+    ne_set_localaddr(sess, ia);
+
+    ONREQ(any_request(sess, "/foo"));
+
+    ne_session_destroy(sess);
+    ne_iaddr_free(ia);
+
+    return reap_server();
+}
+
+/* Regression in 0.27.0, ne_set_progress(sess, NULL, NULL) should
+ * register the progress callback. */
+static int dereg_progress(void)
+{
+    ne_session *sess;
+
+    CALL(make_session(&sess, single_serve_string, 
+                      RESP200 TE_CHUNKED "\r\n" ABCDE_CHUNKS));
+
+    ne_set_progress(sess, NULL, NULL);
+
+    ONREQ(any_request(sess, "/foo"));
+
+    ne_session_destroy(sess);
+    
+    return await_server();    
+}
+
+/* TODO: test that ne_set_notifier(, NULL, NULL) DTRT too. */
+
 ne_test tests[] = {
     T(lookup_localhost),
     T(single_get_clength),
@@ -2080,8 +2164,6 @@ ne_test tests[] = {
     T(no_body_204),
     T(no_body_304),
     T(no_body_HEAD),
-    T(no_body_empty_clength),
-    T(no_body_bad_clength),
     T(no_headers),
     T(chunks),
     T(te_header),
@@ -2097,6 +2179,7 @@ ne_test tests[] = {
     T(persist_http11),
     T(persist_chunked),
     T(persist_http10),
+    T(persist_proxy_http10),
     T(persist_timeout),
     T(no_persist_http10),
     T(ptimeout_eof),
@@ -2157,5 +2240,7 @@ ne_test tests[] = {
     T(icy_protocol),
     T(status),
     T(status_chunked),
+    T(local_addr),
+    T(dereg_progress),
     T(NULL)
 };
